InterpreterPlugin subclass: #XDisplayControlPlugin
	instanceVariableNames: 'osprocessSandboxSecurity '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcess'!

!XDisplayControlPlugin commentStamp: 'dtl 8/3/2003 18:20' prior: 0!
XDisplayControlPlugin provides primitives to control connection to an X display, including disconnecting from the X display to become "headless", and reconnecting to another X display.

This plugin currently works with the Unix VM, but is intended to be extensible for other platforms that support X, and possibly for other external display surfaces.
!


!XDisplayControlPlugin methodsFor: 'initialize-release' stamp: 'dtl 8/3/2003 20:20'!
initialiseModule

	self export: true.
	osprocessSandboxSecurity _ -1.
! !

!XDisplayControlPlugin methodsFor: 'initialize-release' stamp: 'dtl 8/6/2003 05:16'!
shutdownModule

	self export: true.
! !

!XDisplayControlPlugin methodsFor: 'primitives - display' stamp: 'dtl 8/5/2003 21:41'!
primitiveCanConnectToDisplay
	"Open and close a connection to displayName. It the connection was successfully
	opened, answer true; otherwise false. This is intended to check for the ability
	to open an X display prior to actually making the attempt."

	| name namePtr d |
	self export: true.
	self var: 'namePtr' declareC: 'char * namePtr'.
	self var: 'nameBuffer' declareC: 'static char nameBuffer[501]'.
	self var: 'd' declareC: 'Display *d'.

	"Do not allow this if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; push: interpreterProxy falseObject]
		ifFalse:
			[name _ interpreterProxy stackObjectValue: 0.
			namePtr _ self transientCStringFromString: name.
			d _ self XOpenDisplay: namePtr.
			(d == 0)
				ifTrue:	"Failed to make connection to server, answer false"
					[interpreterProxy pop: 2; push: interpreterProxy falseObject]
				ifFalse:	"Successfully opened connection; close it and answer true"
					[self XCloseDisplay: d.
					interpreterProxy pop: 2; push: interpreterProxy trueObject]]
! !

!XDisplayControlPlugin methodsFor: 'primitives - display' stamp: 'dtl 8/6/2003 06:20'!
primitiveDisconnectDisplay
	"Call an internal function which will disconnect the X display session. The actual
	Squeak window on the X server is not effected, but this instance of Squeak will
	not have any further interaction with it."

	self export: true.
	"Do not allow this if running in secure mode"
	(self sandboxSecurity == 1) ifFalse: [self forgetXDisplay]
! !

!XDisplayControlPlugin methodsFor: 'primitives - display' stamp: 'dtl 8/5/2003 21:38'!
primitiveFlushDisplay
	"Call an internal function to synchronize output to the X display."

	self export: true.
	self synchronizeXDisplay
! !

!XDisplayControlPlugin methodsFor: 'primitives - display' stamp: 'dtl 8/3/2003 18:21'!
primitiveGetDisplayName
	"Answer a string containing the name for the X display, or nil if the display was opened
	using the $DISPLAY environment variable. This answers the name of the X display as of
	the time it was last opened, which may be different from the current setting of $DISPLAY."

	| displayName |
	self export: true.
	self var: 'displayName' declareC: 'extern char *displayName'.
	((displayName == 0)
		ifTrue: [interpreterProxy pop: 1; push: (interpreterProxy nilObject)]
		ifFalse: [interpreterProxy pop: 1; push: (self stringFromCString: displayName)])
! !

!XDisplayControlPlugin methodsFor: 'primitives - display' stamp: 'dtl 8/3/2003 18:23'!
primitiveIsConnectedToDisplay
	"Answer true if VM is currently connected to an X server."

	| isConnectedToXServer |
	self var: 'isConnectedToXServer' declareC: 'extern int isConnectedToXServer'.
	self export: true.
	(isConnectedToXServer ~= 0)
		ifTrue: [interpreterProxy pop: 1; push: (interpreterProxy trueObject)]
		ifFalse: [interpreterProxy pop: 1; push: (interpreterProxy falseObject)]
! !

!XDisplayControlPlugin methodsFor: 'primitives - display' stamp: 'dtl 8/6/2003 06:21'!
primitiveKillDisplay
	"Call an internal function to disconnect the X display session and destroy
	the Squeak window on the X display."

	self export: true.
	"Do not allow this if running in secure mode"
	(self sandboxSecurity == 1) ifFalse: [self disconnectXDisplay]
! !

!XDisplayControlPlugin methodsFor: 'primitives - display' stamp: 'dtl 8/5/2003 21:44'!
primitiveOpenDisplay
	"Call an internal function which will open the X display session."

	self export: true.
	"Do not allow this if running in secure mode"
	(self sandboxSecurity == 1) ifFalse: [self openXDisplay]
! !

!XDisplayControlPlugin methodsFor: 'primitives - display' stamp: 'dtl 8/3/2003 18:22'!
primitiveSetDisplayName
	"Set the name for the X display for use in the next call to primitiveOpenXDisplay. Expects
	one parameter which must be either a String or nil."

	| displayName name namePtr nameBuffer |
	self export: true.
	self var: 'namePtr' declareC: 'char * namePtr'.
	self var: 'nameBuffer' declareC: 'static char nameBuffer[501]'.
	self var: 'displayName' declareC: 'extern char *displayName'.
	"Do not allow this if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[name _ interpreterProxy stackObjectValue: 0.
			(name == interpreterProxy nilObject)
				ifTrue:
					[displayName _ 0]
				ifFalse:
					[namePtr _ self transientCStringFromString: name.
					self cCode: 'strncpy(nameBuffer, namePtr, 500)'.
					nameBuffer at: 500 put: 0.
					displayName _ nameBuffer].
			interpreterProxy pop: 1]
! !

!XDisplayControlPlugin methodsFor: 'primitives - version testing' stamp: 'dtl 2/21/2002 16:29'!
primitiveModuleName
	"Answer a string containing the module name string for this plugin."

	| s |
	self export: true.
	self var: 's' declareC: 'char *s'.
	self cCode: 's= (char *)moduleName' inSmalltalk: [s _ moduleName]. "Prevent compiler warning"
	interpreterProxy pop: 1; push: (self stringFromCString: s)
! !

!XDisplayControlPlugin methodsFor: 'primitives - version testing' stamp: 'dtl 3/1/2002 16:56'!
primitiveVersionString
	"Answer a string containing the version string for this plugin."

	self export: true.
	interpreterProxy pop: 1; push: (self stringFromCString: self versionString)
! !

!XDisplayControlPlugin methodsFor: 'version identification' stamp: 'dtl 8/4/2003 18:31'!
versionString
	"Answer a string containing the version string for this plugin. Handle MNU
	errors, which can occur if class InterpreterPlugin has been removed from
	the system.

	Important: When this method is changed, the class side method must also be
	changed to match."

	| version |
	self returnTypeC: 'char *'.
	self var: 'version' declareC: 'static char version[]= "1.0"'.
	^ self cCode: 'version' inSmalltalk: [self class versionString]
! !

!XDisplayControlPlugin methodsFor: 'private' stamp: 'dtl 8/3/2003 20:00'!
stringFromCString: aCString
	"Answer a new String copied from a null-terminated C string.
	Caution: This may invoke the garbage collector."

	| len newString stringPtr |
	self var: 'aCString' declareC: 'char *aCString'.
	self var: 'stringPtr' declareC: 'char *stringPtr'.
	len _ self cCode: 'strlen(aCString)' inSmalltalk: [0].
	newString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len.
	stringPtr _ interpreterProxy arrayValueOf: newString.
	self cCode: '(char *)strncpy(stringPtr, aCString, len)'.
	^ newString

! !

!XDisplayControlPlugin methodsFor: 'private' stamp: 'dtl 8/3/2003 20:14'!
transientCStringFromString: aString
	"Answer a new null-terminated C string copied from aString.
	The string is allocated in object memory, and will be moved
	without warning by the garbage collector. Any C pointer
	reference the the result is valid only until the garbage
	collector next runs. Therefore, this method should only be used
	within a single primitive in a section of code in which the
	garbage collector is guaranteed not to run. Note also that
	this method may itself invoke the garbage collector prior
	to allocating the new C string.

	Warning: The result of this method will be invalidated by the
	next garbage collection, including a GC triggered by creation
	of a new object within a primitive. Do not call this method
	twice to obtain two string pointers."

	| len stringPtr newString cString |
	self returnTypeC: 'char *'.
	self var: 'stringPtr' declareC: 'char *stringPtr'.
	self var: 'cString' declareC: 'char *cString'.
	len _ interpreterProxy sizeOfSTArrayFromCPrimitive: (interpreterProxy arrayValueOf: aString).
	"Allocate space for a null terminated C string."
	interpreterProxy pushRemappableOop: aString.
	newString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len + 1.
	stringPtr _ interpreterProxy arrayValueOf: interpreterProxy popRemappableOop.
	cString _ interpreterProxy arrayValueOf: newString.		"Point to the actual C string."
	self cCode: '(char *)strncpy(cString, stringPtr, len)'.		"Make a copy of the string."
	cString at: (len) put: 0.									"Null terminate the C string."
	^ cString
! !

!XDisplayControlPlugin methodsFor: 'private - security' stamp: 'dtl 3/3/2002 08:44'!
sandboxSecurity
	"Answer 1 if running in secure mode, else 0. The osprocessSandboxSecurity
	variable is initialized to -1. On the first call to this method, set its value to
	either 0 (user has full access to the plugin) or 1 (user is not permitted to do
	dangerous things)."

	(osprocessSandboxSecurity < 0) ifTrue: [osprocessSandboxSecurity _ self securityHeurisitic].
	^ osprocessSandboxSecurity
! !

!XDisplayControlPlugin methodsFor: 'private - security' stamp: 'dtl 3/3/2002 09:29'!
securityHeurisitic
	"Answer 0 to permit full access to OSProcess functions, or 1 if access should be
	restricted for dangerous functions. The rules are:
		- If the security plugin is not present, grant full access
		- If the security plugin can be loaded, restrict access unless user has all
		  of secCanWriteImage, secHasFileAccess and secHasSocketAccess"

	"FIXME: This function has not been tested. -dtl"

	| sCWIfn sHFAfn sHSAfn canWriteImage hasFileAccess hasSocketAccess |
	"If the security plugin can be loaded, use it to check. If not, assume it's ok"

	sCWIfn _ interpreterProxy ioLoadFunction: 'secCanWriteImage' From: 'SecurityPlugin'.
	(sCWIfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	canWriteImage _ self cCode: ' ((int (*) (void)) sCWIfn)()'
				inSmalltalk: [1].
	sHFAfn _ interpreterProxy ioLoadFunction: 'secHasFileAccess' From: 'SecurityPlugin'.
	(sHFAfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	hasFileAccess _ self cCode: ' ((int (*) (void)) sHFAfn)()'
				inSmalltalk: [1].
	sHSAfn _ interpreterProxy ioLoadFunction: 'secHasSocketAccess' From: 'SecurityPlugin'.
	(sHSAfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	hasSocketAccess _ self cCode: ' ((int (*) (void)) sHSAfn)()'
				inSmalltalk: [1].
	((canWriteImage and: [hasFileAccess]) and: [hasSocketAccess])
		ifTrue: [^ 0 "Grant full access"]
		ifFalse: [^ 1 "Restricted access"]
! !


!XDisplayControlPlugin class methodsFor: 'translation' stamp: 'dtl 8/3/2003 20:19'!
declareCVarsIn: cg

	cg addHeaderFile: '<X11/Xlib.h>'.
	cg var: 'osprocessSandboxSecurity' declareC: 'int osprocessSandboxSecurity'.

! !

!XDisplayControlPlugin class methodsFor: 'translation' stamp: 'dtl 8/3/2003 18:13'!
moduleName

	^ 'XDisplayControlPlugin'! !

!XDisplayControlPlugin class methodsFor: 'version identification' stamp: 'dtl 8/4/2003 18:23'!
versionString
	"Important: When this method is changed, the instance side method must also be
	changed to match."

	"self versionString"

	^ '1.0'! !


!XDisplayControlPlugin reorganize!
('initialize-release' initialiseModule shutdownModule)
('primitives - display' primitiveCanConnectToDisplay primitiveDisconnectDisplay primitiveFlushDisplay primitiveGetDisplayName primitiveIsConnectedToDisplay primitiveKillDisplay primitiveOpenDisplay primitiveSetDisplayName)
('primitives - version testing' primitiveModuleName primitiveVersionString)
('version identification' versionString)
('private' stringFromCString: transientCStringFromString:)
('private - security' sandboxSecurity securityHeurisitic)
!

